@@ -13,43 +13,52 @@ logger = logging.getLogger(__name__)
 
 
 class StopTrainingError(Exception):
+    """
+    An exception which can be raised in order to stop a training run early.
+    """
+
+    pass
+
+
+class CallbackMethodNotImplementedError(Exception):
     pass
 
 
 class TrainerCallback(ABC):
+    """
+    The abstract base class to be subclassed when creating new callbacks.
+    """
+
     def on_init_end(self, trainer, **kwargs):
         """
         Event called at the end of trainer initialisation.
         """
         pass
 
-    def on_train_run_begin(self, trainer, **kwargs):
+    def on_training_run_start(self, trainer, **kwargs):
         """
-        Event called at the beginning of training run.
+        Event called at the start of training run.
         """
         pass
 
-    def on_train_run_end(self, trainer, **kwargs):
-        """
-        Event called at the end of training run.
-        """
-        pass
-
-    def on_train_epoch_begin(self, trainer, **kwargs):
+    def on_train_epoch_start(self, trainer, **kwargs):
         """
         Event called at the beginning of a training epoch.
         """
         pass
 
-    def on_train_step_begin(self, trainer, **kwargs):
+    def on_train_step_start(self, trainer, **kwargs):
         """
         Event called at the beginning of a training step.
         """
         pass
 
-    def on_train_step_end(self, trainer, **kwargs):
+    def on_train_step_end(self, trainer, batch, batch_output, **kwargs):
         """
         Event called at the end of a training step.
+
+        :param batch: the current batch of training data
+        :param batch_output: the outputs returned by :meth:`trainer.Trainer.calculate_train_batch_loss`
         """
         pass
 
@@ -59,21 +68,24 @@ class TrainerCallback(ABC):
         """
         pass
 
-    def on_eval_epoch_begin(self, trainer, **kwargs):
+    def on_eval_epoch_start(self, trainer, **kwargs):
         """
         Event called at the beginning of an evaluation epoch.
         """
         pass
 
-    def on_eval_step_begin(self, trainer, **kwargs):
+    def on_eval_step_start(self, trainer, **kwargs):
         """
         Event called at the beginning of a evaluation step.
         """
         pass
 
-    def on_eval_step_end(self, trainer, **kwargs):
+    def on_eval_step_end(self, trainer, batch, batch_output, **kwargs):
         """
         Event called at the end of an evaluation step.
+
+        :param batch: the current batch of evaluation data
+        :param batch_output: the outputs returned by :meth:`trainer.Trainer.calculate_eval_batch_loss`
         """
         pass
 
@@ -83,13 +95,47 @@ class TrainerCallback(ABC):
         """
         pass
 
-    def on_stop_training_error(self, trainer, **kwargs):
+    def on_training_run_epoch_end(self, trainer, **kwargs):
+        """
+        Event called during a training run after both training and evaluation epochs have been completed.
+        """
         pass
 
+    def on_training_run_end(self, trainer, **kwargs):
+        """
+        Event called at the end of training run.
+        """
+        pass
+
+    def on_evaluation_run_start(self, trainer, **kwargs):
+        """
+        Event called at the start of an evaluation run.
+        """
+        pass
+
+    def on_evaluation_run_end(self, trainer, **kwargs):
+        """
+        Event called at the end of an evaluation run.
+        """
+        pass
+
+    def on_stop_training_error(self, trainer, **kwargs):
+        """
+        Event called when a stop training error is raised
+        """
+        pass
+
+    def __getattr__(self, item):
+        try:
+            return super().__getattr__(item)
+        except AttributeError:
+            raise CallbackMethodNotImplementedError
+
 
 class CallbackHandler:
     """
-    Responsible for calling a list of callbacks. This class calls the callbacks in the order that they are given.
+    The :class:`CallbackHandler` is responsible for calling a list of callbacks.
+    This class calls the callbacks in the order that they are given.
     """
 
     def __init__(self, callbacks):
@@ -97,10 +143,20 @@ class CallbackHandler:
         self.add_callbacks(callbacks)
 
     def add_callbacks(self, callbacks):
+        """
+        Add a list of callbacks to the callback handler
+
+        :param callbacks: a list of :class:`TrainerCallback`
+        """
         for cb in callbacks:
             self.add_callback(cb)
 
     def add_callback(self, callback):
+        """
+        Add a callbacks to the callback handler
+
+        :param callback: an instance of a subclass of :class:`TrainerCallback`
+        """
         cb = callback() if isinstance(callback, type) else callback
         cb_class = callback if isinstance(callback, type) else callback.__class__
         if cb_class in [c.__class__ for c in self.callbacks]:
@@ -121,14 +177,33 @@ class CallbackHandler:
         return "\n".join(cb.__class__.__name__ for cb in self.callbacks)
 
     def call_event(self, event, *args, **kwargs):
+        """
+        For each callback which has been registered, sequentially call the method corresponding to the
+        given event.
+
+        :param event: The event corresponding to the method to call on each callback
+        :param args: a list of arguments to be passed to each callback
+        :param kwargs: a list of keyword arguments to be passed to each callback
+        """
         for callback in self.callbacks:
-            getattr(callback, event)(
-                *args,
-                **kwargs,
-            )
+            try:
+                getattr(callback, event)(
+                    *args,
+                    **kwargs,
+                )
+            except CallbackMethodNotImplementedError as e:
+                continue
 
 
 class PrintMetricsCallback(TrainerCallback):
+    """
+    A callback that prints the latest values of any metric which has been updated in
+    the trainer's run history.
+
+    Metrics prefixed with 'train' are printed at the end of a training epoch, all other
+    metrics are printed after evaluation.
+    """
+
     def _print_metrics(self, trainer, metric_names):
         for metric_name in metric_names:
             trainer.print(
@@ -154,10 +229,14 @@ class PrintMetricsCallback(TrainerCallback):
 
 
 class ProgressBarCallback(TrainerCallback):
+    """
+    A callback which visualises the state of each training and evaluation epoch using a progress bar
+    """
+
     def __init__(self):
         self.pbar = None
 
-    def on_train_epoch_begin(self, trainer, **kwargs):
+    def on_train_epoch_start(self, trainer, **kwargs):
         self.pbar = tqdm(
             total=len(trainer._train_dataloader),
             disable=not trainer._accelerator.is_local_main_process,
@@ -170,7 +249,7 @@ class ProgressBarCallback(TrainerCallback):
         self.pbar.close()
         time.sleep(0.01)
 
-    def on_eval_epoch_begin(self, trainer, **kwargs):
+    def on_eval_epoch_start(self, trainer, **kwargs):
         self.pbar = tqdm(
             total=len(trainer._eval_dataloader),
             disable=not trainer._accelerator.is_local_main_process,
@@ -185,45 +264,65 @@ class ProgressBarCallback(TrainerCallback):
 
 
 class PrintProgressCallback(TrainerCallback):
-    def on_train_run_begin(self, trainer, **kwargs):
+    """
+    A callback which prints a message at the start and end of a run,
+    as well as at the start of each epoch.
+    """
+
+    def on_training_run_start(self, trainer, **kwargs):
         trainer.print("\nStarting training run")
 
-    def on_train_epoch_begin(self, trainer, **kwargs):
+    def on_train_epoch_start(self, trainer, **kwargs):
         trainer.print(f"\nStarting epoch {trainer.run_history.current_epoch}")
         time.sleep(0.01)
 
-    def on_train_run_end(self, trainer, **kwargs):
+    def on_training_run_end(self, trainer, **kwargs):
         trainer.print("Finishing training run")
 
+    def on_evaluation_run_start(self, trainer, **kwargs):
+        trainer.print("\nStarting evaluation run")
+
+    def on_evaluation_run_end(self, trainer, **kwargs):
+        trainer.print("Finishing evaluation run")
+
 
 class SaveBestModelCallback(TrainerCallback):
+    """
+    A callback which saves the best model during a training run, according to a given metric.
+    The best model weights are loaded at the end of the training run.
+    """
+
     def __init__(
         self,
-        save_dir=None,
+        save_path="best_model.pt",
         watch_metric="eval_loss_epoch",
-        greater_is_better=False,
-        reset_on_train=True,
+        greater_is_better: bool = False,
+        reset_on_train: bool = True,
     ):
+        """
+
+        :param save_path: The path to save the checkpoint to. This should end in ``.pt``.
+        :param watch_metric: the metric used to compare model performance. This should be accessible from the trainer's run history.
+        :param greater_is_better: whether an increase in the ``watch_metric`` should be interpreted as the model performing better.
+        :param reset_on_train: whether to reset the best metric on subsequent training runs. If ``True``, only the metrics observed during the current training run will be compared.
+        """
         self.watch_metric = watch_metric
         self.greater_is_better = greater_is_better
         self.operator = np.greater if self.greater_is_better else np.less
         self.best_metric = None
-        self.save_dir = save_dir
+        self.save_path = save_path
         self.reset_on_train = reset_on_train
 
-    def on_train_run_begin(self, args, **kwargs):
-        if self.save_dir is None:
-            self.save_dir = "model.pt"
-
+    def on_training_run_start(self, args, **kwargs):
         if self.reset_on_train:
             self.best_metric = None
 
-    def on_eval_epoch_end(self, trainer, **kwargs):
+    def on_training_run_epoch_end(self, trainer, **kwargs):
         current_metric = trainer.run_history.get_latest_metric(self.watch_metric)
         if self.best_metric is None:
             self.best_metric = current_metric
-            trainer.save_model(
-                save_path=self.save_dir,
+            trainer.save_checkpoint(
+                save_path=self.save_path,
                 checkpoint_kwargs={self.watch_metric: self.best_metric},
             )
         else:
@@ -231,27 +330,39 @@ class SaveBestModelCallback(TrainerCallback):
 
             if is_improvement:
                 self.best_metric = current_metric
-                trainer.save_model(
-                    save_path=self.save_dir,
+                trainer.save_checkpoint(
+                    save_path=self.save_path,
                     checkpoint_kwargs={"loss": self.best_metric},
                 )
 
-    def on_train_run_end(self, trainer, **kwargs):
+    def on_training_run_end(self, trainer, **kwargs):
         trainer.print(
             f"Loading checkpoint with {self.watch_metric}: {self.best_metric}"
         )
-        trainer.load_checkpoint(self.save_dir)
+        trainer.load_checkpoint(self.save_path)
 
 
 class EarlyStoppingCallback(TrainerCallback):
+    """
+    A callback which stops training early if progress is not being observed.
+    """
+
     def __init__(
         self,
         early_stopping_patience: int = 1,
-        early_stopping_threshold: Optional[float] = 0.01,
+        early_stopping_threshold: float = 0.01,
         watch_metric="eval_loss_epoch",
-        greater_is_better=False,
-        reset_on_train=True,
+        greater_is_better: bool = False,
+        reset_on_train: bool = True,
     ):
+        """
+
+        :param early_stopping_patience: the number of epochs with no improvement after which training will be stopped.
+        :param early_stopping_threshold: the minimum change in the ``watch_metric`` to qualify as an improvement, i.e. an absolute change of less than this threshold, will count as no improvement.
+        :param watch_metric: the metric used to compare model performance. This should be accessible from the trainer's run history.
+        :param greater_is_better: whether an increase in the ``watch_metric`` should be interpreted as the model performing better.
+        :param reset_on_train: whether to reset the best metric on subsequent training runs. If ``True``, only the metrics observed during the current training run will be compared.
+        """
         self.early_stopping_patience = early_stopping_patience
         self.early_stopping_threshold = early_stopping_threshold
         self.watch_metric = watch_metric
@@ -261,12 +372,12 @@ class EarlyStoppingCallback(TrainerCallback):
         self.operator = np.greater if self.greater_is_better else np.less
         self.reset_on_train = reset_on_train
 
-    def on_train_run_begin(self, args, **kwargs):
+    def on_training_run_start(self, args, **kwargs):
         if self.reset_on_train:
             self.best_metric = None
             self.early_stopping_patience_counter = 0
 
-    def on_eval_epoch_end(self, trainer, **kwargs):
+    def on_training_run_epoch_end(self, trainer, **kwargs):
         current_metric = trainer.run_history.get_latest_metric(self.watch_metric)
         if self.best_metric is None:
             self.best_metric = current_metric
@@ -289,7 +400,10 @@ class EarlyStoppingCallback(TrainerCallback):
 
 
 class TerminateOnNaNCallback(TrainerCallback):
-    """A callback that terminates training if loss is NaN."""
+    """
+    A callback that terminates the training run if a ``NaN`` loss is observed during either training or
+    evaluation.
+    """
 
     def __init__(self):
         self.triggered = False
@@ -307,6 +421,6 @@ class TerminateOnNaNCallback(TrainerCallback):
     def on_eval_step_end(self, trainer, batch_output, **kwargs):
         self.check_for_nan_after_batch(batch_output, step="validation")
 
-    def on_train_run_end(self, trainer, **kwargs):
+    def on_training_run_end(self, trainer, **kwargs):
         if self.triggered:
             sys.exit("Exiting due to NaN loss")
